"""All abstract data types to interact with Raindrop's API."""
from __future__ import annotations

import datetime
import enum
import json
from pathlib import Path
from typing import Any, Optional, Sequence

from pydantic import (
    BaseModel,
    Field,
    PositiveInt,
    EmailStr,
    NonNegativeInt,
    root_validator,
    HttpUrl,
)

from .api import API

__all__ = [
    "Access",
    "AccessLevel",
    "BrokenLevel",
    "Collection",
    "CollectionRef",
    "FontColor",
    "Group",
    "Raindrop",
    "RaindropType",
    "Tag",
    "User",
    "UserFiles",
    "UserRef",
    "View",
]

# Base URL
URL = "https://api.raindrop.io/rest/v1/{path}"


class AccessLevel(enum.IntEnum):
    """Map the Access levels defined by Raindrop's API."""

    readonly = 1
    collaborator_read = 2
    collaborator_write = 3
    owner = 4


class CacheStatus(enum.Enum):
    """Represents the various states the cache of a Raindrop might be in."""

    ready = "ready"
    retry = "retry"
    failed = "failed"
    invalid_origin = "invalid-origin"
    invalid_timeout = "invalid-timeout"
    invalid_size = "invalid-size"


class View(enum.Enum):
    """Map the names of the views for Raindrop's API."""

    list = "list"
    simple = "simple"
    grid = "grid"
    masonry = "masonry"


class CollectionRef(BaseModel):
    """Represents a *reference* to a Raindrop Collection (almost a TypeVar of id: int)."""

    id: int = Field(None, alias="$id")


# We define the 3 "system" collections in the Raindrop environment:
CollectionRef.All = CollectionRef(**{"$id": 0})
CollectionRef.Trash = CollectionRef(**{"$id": -99})
CollectionRef.Unsorted = CollectionRef(**{"$id": -1})


class UserRef(BaseModel):
    """Represents a *reference* to :class:`User` object."""

    id: int = Field(None, alias="$id")
    ref: str = Field(None, alias="$user")


class Access(BaseModel):
    """Represents Access control of Collections."""

    level: AccessLevel
    draggable: bool


class Collection(BaseModel):
    """Represents a concrete Raindrop Collection."""

    id: int = Field(None, alias="_id")
    title: str
    user: UserRef

    access: Access | None
    collaborators: list[Any] | None = []
    color: str | None = None
    count: NonNegativeInt
    cover: list[str] | None = []
    created: datetime.datetime | None
    expanded: bool = False
    lastUpdate: datetime.datetime | None
    parent: int | None  # Id of parent collection
    public: bool | None
    sort: NonNegativeInt | None
    view: View | None

    @classmethod
    def get_roots(cls, api: API) -> Sequence[Collection]:
        """Get "root" Raindrop collections."""
        ret = api.get(URL.format(path="collections"))
        items = ret.json()["items"]
        return [cls(**item) for item in items]

    @classmethod
    def get_childrens(cls, api: API) -> Sequence[Collection]:
        """Get the "child" Raindrop collections (ie. all below root level)."""
        ret = api.get(URL.format(path="collections/childrens"))
        items = ret.json()["items"]
        return [cls(**item) for item in items]

    @classmethod
    def get_collections(cls, api: API) -> Sequence[Collection]:
        """Query *ALL* collections. Wrapper on get_roots and get_childrens.

        (hiding the distinction between root/child collections)
        """
        return cls.get_roots(api) + cls.get_childrens(api)

    @classmethod
    def get(cls, api: API, id: int) -> Collection:
        """Return a Raindrop collection based on it's id."""
        url = URL.format(path=f"collection/{id}")
        item = api.get(url).json()["item"]
        return cls(**item)

    @classmethod
    def create(
        cls,
        api: API,
        view: Optional[View] = None,
        title: Optional[str] = None,
        sort: Optional[int] = None,
        public: Optional[bool] = None,
        parent: Optional[int] = None,
        cover: Optional[Sequence[str]] = None,
    ) -> Collection:
        """Create a new Raindrop collection."""
        args: dict[str, Any] = {}
        if view is not None:
            args["view"] = view
        if title is not None:
            args["title"] = title
        if sort is not None:
            args["sort"] = sort
        if public is not None:
            args["public"] = public
        if parent is not None:
            args["parent"] = parent
        if cover is not None:
            args["cover"] = cover

        url = URL.format(path="collection")
        item = api.post(url, json=args).json()["item"]
        return cls(**item)

    @classmethod
    def update(
        cls,
        api: API,
        id: int,
        expanded: Optional[bool] = None,
        view: Optional[View] = None,
        title: Optional[str] = None,
        sort: Optional[int] = None,
        public: Optional[bool] = None,
        parent: Optional[int] = None,
        cover: Optional[Sequence[str]] = None,
    ) -> Collection:
        """Update an existing Raindrop collection with any of the attribute values provided."""
        args: dict[str, Any] = {}
        for attr in ["expanded", "view", "title", "sort", "public", "parent", "cover"]:
            if (value := locals().get(attr)) is not None:
                args[attr] = value
        url = URL.format(path=f"collection/{id}")
        item = api.put(url, json=args).json()["item"]
        return cls(**item)

    @classmethod
    def remove(cls, api: API, id: int) -> None:
        """Remove/delete a Raindrop collection."""
        api.delete(URL.format(path=f"collection/{id}"), json={})

    @classmethod
    def get_or_create(cls, api: API, title: str) -> Collection:
        """Get a Raindrop collection based on it's *title*, if it doesn't exist, create it.

        Return the ID associated with the collection with specified collection title (this
        doesn't seem to be a supported method of the Raindrop API directly). If collection
        is not found, add it!.
        """
        # Note: Since
        for collection in Collection.get_roots(api):
            if title.casefold() == collection.title.casefold():
                return collection

        # Doesn't exist, create it!
        return Collection.create(api, title=title)


class BrokenLevel(enum.Enum):
    """Enumerate user levels."""

    basic = "basic"
    default = "default"
    strict = "strict"
    off = "off"


class FontColor(enum.Enum):
    """Enumerate user display themes available."""

    sunset = "sunset"
    night = "night"


class RaindropSort(enum.Enum):
    """Enumerate user default sort options available."""

    title_up = "title"
    title_dn = "-title"
    domain_up = "domain"
    domain_dn = "-domain"
    last_update_up = "+lastUpdate"
    last_update_dn = "-lastUpdate"


class UserConfig(BaseModel):
    """Sub-model defining a Raindrop user's configuration."""

    broken_level: BrokenLevel = None
    font_color: FontColor | None = None
    font_size: int | None = None
    last_collection: CollectionRef | None = None
    raindrops_sort: RaindropSort | None = None
    raindrops_view: View | None = None


class Group(BaseModel):
    """Sub-model defining a Raindrop user group."""

    title: str
    hidden: bool
    sort: NonNegativeInt
    collectionids: list[int] = Field(None, alias="collections")


class UserFiles(BaseModel):
    """Sub-model defining a file associated with a user (?)."""

    used: int
    size: PositiveInt
    lastCheckPoint: datetime.datetime


class User(BaseModel):
    """Raindrop User model."""

    id: int = Field(None, alias="_id")
    email: EmailStr
    email_MD5: str | None = None
    files: UserFiles
    fullName: str
    groups: list[Group]
    password: bool
    pro: bool
    proExpire: datetime.datetime | None = None
    registered: datetime.datetime

    @classmethod
    def get(cls, api: API) -> User:
        """Get all the information about the Raindrop user associated with the API token."""
        user = api.get(URL.format(path="user")).json()["user"]
        return cls(**user)


class SystemCollection(BaseModel):
    """Raindrop System Collection model, ie. collections for Unsorted, Trash and 'All'."""

    id: int = Field(None, alias="_id")
    count: NonNegativeInt
    title: str | None

    @root_validator(pre=False)
    def _map_systemcollection_id_to_title(cls, values):
        """Map the hard-coded id's of the System Collections to the descriptions used on the UI."""
        _titles = {
            CollectionRef.Unsorted.id: "Unsorted",
            CollectionRef.All.id: "All",
            CollectionRef.Trash.id: "Trash",
        }
        values["title"] = _titles.get(values["id"])
        return values

    @classmethod
    def get_status(cls, api: API) -> User:
        """Get the count of Raindrops across all 3 'system' collections."""
        items = api.get(URL.format(path="user/stats")).json()["items"]
        return [cls(**item) for item in items]


class RaindropType(enum.Enum):
    """Map the types of Raindrop bookmarks possible (ie. what type of content they hold)."""

    link = "link"
    article = "article"
    image = "image"
    video = "video"
    document = "document"
    audio = "audio"


class File(BaseModel):
    """Represents the attributes associated with a file within a document-based Raindrop."""

    name: str
    size: PositiveInt
    type: str


class Cache(BaseModel):
    """Represents the cache information of Raindrop."""

    status: CacheStatus
    size: PositiveInt | None = None
    created: datetime.datetime | None = None


class Raindrop(BaseModel):
    """Core class of a Raindrop bookmark 'item'."""

    # "Main" fields (per https://developer.raindrop.io/v1/raindrops)
    id: int = Field(None, alias="_id")
    collection: CollectionRef | Collection = CollectionRef.Unsorted

    cover: str | None
    created: datetime.datetime | None
    domain: str | None
    excerpt: str | None  # aka 'Description' on the Raindrop UI.
    lastUpdate: datetime.datetime | None
    link: HttpUrl | None
    media: list[dict[str, Any]] | None
    tags: list[str] | None
    title: str | None
    type: RaindropType | None
    user: UserRef | None

    # "Other" fields:
    broken: bool | None
    cache: Cache | None
    file: File | None
    important: bool | None  # aka marked as Favorite.

    @classmethod
    def get(cls, api: API, id: int) -> Raindrop:
        """Return a Raindrop bookmark based on it's id."""
        item = api.get(URL.format(path=f"{id}")).json()["item"]
        return cls(**item)

    @classmethod
    def create_link(
        cls,
        api: API,
        link: str,
        collection: Optional[Collection | CollectionRef, int] = None,
        cover: Optional[str] = None,
        created: Optional[datetime.datetime] = None,
        excerpt: Optional[str] = None,
        html: Optional[str] = None,
        important: Optional[bool] = None,
        lastUpdate: Optional[datetime.datetime] = None,
        media: Optional[Sequence[dict[str, Any]]] = None,
        order: Optional[int] = None,
        pleaseParse: bool = False,  # If set, asks API to automatically parse metadata in the background
        tags: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
        type: Optional[str] = None,
    ) -> Raindrop:
        """Create a new link-type Raindrop bookmark."""
        # NOTE: We have small code style conflict here between Vulture
        # and Ruff, specifically:

        # Vulture will report all the optional variables above as
        # "unused". This is clearly technically true as their only
        # appearance below is as a string to the arg below from which
        # we take the actual value of the inbound argument from
        # locals(), instead of referring to the argument explicitly.
        #
        # However, converting all of these to a (rather lengthy set)
        # of simple statements like:
        #
        # if <arg> is not None:
        #     args[<arg>] = <arg>
        #
        # keeps Vulture happy but now trips the McCabe complexity
        # metric(!) implemented in our Ruff pre-commit pass due to the
        # number of conditionals in a single method!
        #
        # While, we /could/, convert to kwargs, we lose the documentation
        # and verbosity available with explicit arguments.
        #
        # Thus, for now, we *leave* vulture out of pre-commit and
        # simply run it manually as necessary (through our justfile)

        # Setup the args that will be passed to the underlying
        # Raindrop API, only link is absolutely required, rest are
        # optional!
        args: dict[str, Any] = {"link": link}

        if pleaseParse:
            args["pleaseParse"] = {}

        for attr in [
            "cover",
            "created",
            "excerpt",
            "html",
            "important",
            "lastUpdate",
            "media",
            "order",
            "tags",
            "title",
            "type",
        ]:
            if (value := locals().get(attr)) is not None:
                args[attr] = value

        if collection is not None:
            # <collection> arg could be *either* an actual collection
            # or simply an int collection "id" already, handle either:
            if isinstance(collection, (Collection, CollectionRef)):
                args["collection"] = {"$id": collection.id}
            else:
                args["collection"] = {"$id": collection}

        url = URL.format(path="raindrop")
        item = api.post(url, json=args).json()["item"]
        return cls(**item)

    @classmethod
    def create_file(
        cls,
        api: API,
        path: Path,
        content_type: str,
        collection: int = -1,
        tags: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
    ) -> Raindrop:
        """Create a new file-based Raindrop bookmark."""
        url = URL.format(path="raindrop/file")

        # Structure here confirmed through communication with RustemM
        # on 2022-11-29 and his subsequent update to API docs.
        data = {"collectionId": str(collection)}
        files = {"file": (path.name, open(path, "rb"), content_type)}
        item = api.put_file(url, path, data, files).json()["item"]
        raindrop = cls(**item)

        # The Raindrop API's "Create Raindrop From File" does not
        # allow us to set other attributes, thus, we need to check if
        # any of the possible attributes need to be set and do so
        # explicitly with another call:
        args: dict[str, Any] = {}
        if title is not None:
            args["title"] = title
        if tags is not None:
            args["tags"] = tags
        if args:
            url = URL.format(path=f"raindrop/{raindrop.id}")
            item = api.put(url, json=args).json()["item"]
            return cls(**item)
        else:
            return raindrop

    @classmethod
    def update(
        cls,
        api: API,
        id: int,
        collection: Optional[Collection | CollectionRef, int] = None,
        cover: Optional[str] = None,
        created: Optional[datetime.datetime] = None,
        excerpt: Optional[str] = None,
        html: Optional[str] = None,
        important: Optional[bool] = None,
        lastUpdate: Optional[datetime.datetime] = None,
        link: Optional[str] = None,
        media: Optional[Sequence[dict[str, Any]]] = None,
        order: Optional[int] = None,
        pleaseParse: Optional[bool] = False,
        tags: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
        type: Optional[str] = None,
    ) -> Raindrop:
        """Update an existing Raindrop bookmark with any of the attribute values provided."""
        # Setup the args that will be passed to the underlying Raindrop API
        # optional!
        args: dict[str, Any] = {}

        if pleaseParse:
            args["pleaseParse"] = {}

        for attr in [
            "cover",
            "created",
            "excerpt",
            "html",
            "important",
            "lastUpdate",
            "link",
            "media",
            "order",
            "tags",
            "title",
            "type",
        ]:
            if (value := locals().get(attr)) is not None:
                args[attr] = value

        if collection is not None:
            # <collection> arg could be *either* an actual collection
            # or simply an int collection "id" already, handle either:
            if isinstance(collection, (Collection, CollectionRef)):
                args["collection"] = collection.id
            else:
                args["collection"] = collection

        url = URL.format(path=f"raindrop/{id}")
        item = api.put(url, json=args).json()["item"]
        return cls(**item)

    @classmethod
    def remove(cls, api: API, id: int) -> None:
        """Remove/delete a Raindrop bookmark."""
        api.delete(URL.format(path=f"raindrop/{id}"), json={})

    @classmethod
    def search_paged(
        cls,
        api: API,
        collection: CollectionRef = CollectionRef.All,
        page: int = 0,
        perpage: int = 50,
        word: Optional[str] = None,
        tag: Optional[str] = None,
        important: Optional[bool] = None,
    ) -> list[Raindrop]:
        """Lower-level search for bookmarks on a "paged" basis.

        In the specified collection with key word, tag or importance parameters.
        """
        args: list[dict[str, Any]] = []
        if word is not None:
            args.append({"key": "word", "val": word})
        if tag is not None:
            args.append({"key": "tag", "val": tag})
        if important is not None:
            args.append({"key": "important", "val": important})

        params = {"search": json.dumps(args), "perpage": perpage, "page": page}
        url = URL.format(path=f"raindrops/{collection.id}")
        results = api.get(url, params=params).json()
        return [cls(**item) for item in results["items"]]

    @classmethod
    def search(
        cls,
        api: API,
        collection: CollectionRef = CollectionRef.All,
        word: Optional[str] = None,
        tag: Optional[str] = None,
        important: Optional[bool] = None,
    ) -> list[Raindrop]:
        """Search for Raindrops.

        Essentially, a simple wrapper over the paged version above
        """
        page = 0
        results = list()
        while raindrops := Raindrop.search_paged(
            api,
            collection,
            page=page,
            word=word,
            tag=tag,
            important=important,
        ):
            results.extend(raindrops)
            page += 1
        return results


class Tag(BaseModel):
    """Represents existing Tags, either all or just a specific collection."""

    tag: str = Field(None, alias="_id")
    count: int

    @classmethod
    def get(cls, api: API, collection_id: int = None) -> list[Tag]:
        """Get all the tags currently defined, either in a specific collections or across all collections."""
        url = URL.format(path="tags")
        if collection_id:
            url += "/" + str(collection_id)
        items = api.get(url).json()["items"]
        return [Tag(**item) for item in items]

    @classmethod
    def remove(cls, api: API, tags: Sequence[str]) -> None:
        """Remove/delete one or more tags."""
        api.delete(URL.format(path="tags"), json={})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['raindropiopy',
 'raindropiopy.api',
 'raindropiopy.cli',
 'raindropiopy.cli.commands',
 'raindropiopy.cli.models']

package_data = \
{'': ['*']}

install_requires = \
['email-validator>=1.3.1,<2.0.0',
 'humanize>=4.4.0,<5.0.0',
 'prompt-toolkit>=3.0.36,<4.0.0',
 'pydantic>=1.10.4,<2.0.0',
 'pyfiglet>=0.8.post1,<0.9',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'requests-oauthlib>=1.3.1,<2.0.0',
 'rich>=13.2.0,<14.0.0',
 'tomli>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['raindrop-io-py = raindropiopy.cli.cli:main']}

setup_kwargs = {
    'name': 'raindrop-io-py',
    'version': '0.0.14',
    'description': 'API and terminal-based CLI for Raindrop.io bookmark manager',
    'long_description': '[![version](https://img.shields.io/badge/python-3.10+-green)](https://www.python.org/)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit)](https://github.com/pre-commit/pre-commit)\n[![license](https://img.shields.io/badge/License-MIT-green.svg)](https://github.com/PBorocz/raindrop-io-py/blob/trunk/LICENSE)\n\n\n# Raindrop-IO-py\n\nPython wrapper for the API to the [Raindrop.io](https://raindrop.io) Bookmark Manager as well as a simple command-line interface to perform common operations.\n\n## Background\n\nI wanted to use an existing API for the Raindrop Bookmark Manager ([python-raindropio](https://github.com/atsuoishimoto/python-raindropio)) to perform some bulk operations through a simple command-line interface. However, the API available was incomplete and didn\'t contain any user-interface. Thus, this is a _fork_ and significant extension of [python-raindropio](https://github.com/atsuoishimoto/python-raindropio) (ht [Atsuo Ishimoto](https://github.com/atsuoishimoto)).\n\nThis package includes:\n\n-   An API providing access to the Raindrop environment. For instance: create, update, delete link/file-based Raindrops; create, update delete Raindrop collections, tags etc.\n-   A terminal-based user-interface that both tests the API as well as providing (me) a fast, simple interface to my Raindrop collections.\n\n## Status\n\nAs the API layer is based on a fork of an existing package, it\'s reasonably stable. However, the command-line interface (CLI) is brand new (and lacking tests, ie. "works for me!" ;-).\n\n## Requirements\n\nRequires Python 3.10 or later (well, at least I\'m developing against 3.10.9).\n\n## Install\n\n```shell\n[.venv] python -m pip install raindrop-io-py\n```\n\n## Setup\n\nTo use this package, besides your own account on [Raindrop](https://raindrop.io), you\'ll need to create an `integration app` on the Raindrop.io site from which you can create API token(s).\n\n-   Go to [<https://app.draindrop.api/settings/integrations>](https://app.raindrop.io/settings/integrations) and select `+ create new app`.\n\n-   Give it a descriptive name and then select the app you just created.\n\n-   Select `Create test token` and copy the token provided. Note that the basis for calling it a _test_ token is that it only gives you access to bookmarks within *your own account*. Raindrop allows you to use their API against other people\'s environments using oAuth (see untested/unsupported `flask_oauth.py` file in /examples)\n\n-   Save your token into your environment (we use python-dotenv so a simple .env/.envrc file containing your token should suffice), for example:\n\n```shell\n# If you use direnv or it\'s equivalent, place something like this in a .env file:\nRAINDROP_TOKEN=01234567890-abcdefghf-aSample-API-Token-01234567890-abcdefghf\n\n# Or for bash:\nexport RAINDROP_TOKEN=01234567890-abcdefghf-aSample-API-Token-01234567890-abcdefghf\n\n# Or for fish:\nset -gx RAINDROP_TOKEN 01234567890-abcdefghf-aSample-API-Token-01234567890-abcdefghf\n\n# etc...\n```\n\n## Examples\n\nA full suite of examples are provided in the `examples` directory. Each can be run independently as:\n\n```shell\n[.venv] % python examples/list_collections.py\n```\n\nor a wrapper script is available to run all of them, in logical order with a small wait to be nice to Raindrop\'s API:\n\n```shell\n[.venv] % python examples/RUN_ALL.py\n```\n\n### API Examples\n\nHere are a few examples of API usage (all of these should be able to be executed "as-is"):\n\n#### Display All Collections and **Unsorted** Bookmarks:\n\nThis example shows the intended usage of the API as a context-manager, from which any number of calls can be made:\n\n```python\nimport os\n\nfrom dotenv import load_dotenv\n\nfrom raindropiopy.api import API, Collection, CollectionRef, Raindrop\n\nload_dotenv()\n\nwith API(os.environ["RAINDROP_TOKEN"]) as api:\n\n    print("Current Collections:"\n    for collection in Collection.get_collections(api):\n        print(collection.title)\n\n    print("\\nUnsorted Raindrop Bookmarks:"\n    for item in Raindrop.search(api, collection=CollectionRef.Unsorted):\n        print(item.title)\n```\n\n#### Create a New Raindrop Bookmark to a URL\n\n```python\nimport os\n\nfrom dotenv import load_dotenv\n\nfrom raindropiopy.api import API, Raindrop\n\nload_dotenv()\n\nwith API(os.environ["RAINDROP_TOKEN"]) as api:\n    link, title = "https://www.python.org/", "Our Benevolent Dictator\'s Creation"\n    print(f"Creating Raindrop to: \'{link}\' with title: \'{title}\'...", flush=True, end="")\n    raindrop = Raindrop.create_link(api, link=link, title=title, tags=["abc", "def"])\n    print(f"Done, id={raindrop.id}")\n\n```\n\n(after this has executed, go to your Raindrop.io environment (site or app) and you should see this Raindrop to python.org available)\n\n#### Create a New Raindrop Collection\n\n```python\nimport os\nimport sys\nfrom datetime import datetime\nfrom getpass import getuser\n\nfrom dotenv import load_dotenv\n\nfrom raindropiopy.api import API, Collection\n\nload_dotenv()\n\nwith API(os.environ["RAINDROP_TOKEN"]) as api:\n    title = f"TEST Collection ({getuser()}@{datetime.now():%Y-%m-%dT%H:%M:%S})"\n    print(f"Creating collection: \'{title}\'...", flush=True, end="")\n    collection = Collection.create(api, title=title)\n    print(f"Done, {collection.id=}.")\n```\n\n(after this has executed, go to your Raindrop.io environment (site or app) and you should see this collection available)\n\n## Command-Line Interface Usage\n\n```shell\n[.v\nenv] % raindropiopy\n```\n\nNote: remember to setup `RAINDROP-TOKEN` in your environment!\n\n## Acknowledgments\n\n- [python-raindropio](https://github.com/atsuoishimoto/python-raindropio) from [Atsuo Ishimoto](https://github.com/atsuoishimoto).\n\n## License\n\nThe project is licensed under the MIT License.\n\n## Release History\n\n### Unreleased\n\n### v0.0.14 - 2023-02-09\n\n- FIXED: `Raindrop.cache.size` and `Raindrop.cache.created` attributes are now optional (RaindropIO\'s API doesn\'t always provide them).\n- FIXED: README examples corrected to reflect simpler Raindrop.search call.\n\n### v0.0.13 - 2023-02-07\n\n- CHANGED: Cross-referenced the fields available from the Raindrop API with our API; most available but several optional ones skipped for now.\n- CHANGED: (Internal) Remove dependency on ["jashin"](https://github.com/sojin-project/jashin) library by moving to [pydantic](https://docs.pydantic.dev/) for all Raindrop API models.\n\n### v0.0.12 - 2023-02-06\n\n- CHANGED: (Internal) Move from README.org to README.md to allow PyPI to display project information correctly.\n\n### v0.0.11 - 2023-02-06\n\n- CHANGED: Raindrop search API call is now non-paged (the "paged" version is still available as `Raindrop.search_paged`).\n\n### v0.0.10 - 2023-02-05\n\n- ADDED: Ability to specify raindrop field: Description on a created Raindrop (either file or link-based).\n- ADDED: Ability to re-query existing search results (eg. after changes) and smoothed out post-search interactions.\n\n### v0.0.9 - 2023-02-04\n\n- ADDED: An ability to view, edit and delete raindrops returned from a search.\n- ADDED: A simple `RUN_ALL.py` script to the examples directory to...well, run all the examples in order!\n- CHANGED: The display of raindrops returned from a search to include tags and to only show Collection name if all raindrops are across multiple collections.\n\n### v0.0.8 - 2023-01-25\n\n- CHANGED: Added simple version method in root package:\n\n```python\nfrom raindropiopy import version\nprint(version())\n```\n\n### v0.0.7 - 2023-01-25\n\n- CHANGED: Moved from keeping README in markdown to org file format. Incorporated package\'s ChangeLog into README as well (at the bottom).\n- CHANGED: Added new manage.py release automation capability (internal only, nothing public-facing).\n\n### v0.0.6 - 2023-01-22\n\n- FIXED: CLI autocomplete now works again after adding support for "single-letter" command-shortcuts.\n- ADDED: A set of missing attributes to the Raindrop API model type, eg. file, cache etc. Only attribute still missing is "highlights".\n\n### v0.0.5 - 2023-01-21\n\n- ADDED: Support use of [Vulture](https://github.com/jendrikseipp/vulture) for dead-code analysis (not in pre-commit through due to conflict with ruff\'s McCabe complexity metric)\n- CHANGED: Moved internal module name to match that of package name. Since we couldn\'t use raindroppy as a package name on PyPI due to similarities with existing packages (one of which was for a **crypto** package), we renamed this package to raindrop-io-py. In concert, the internal module is now `raindropiopy`:\n\n```python\nfrom raindroiopy.api import API\n```\n\n- FIXED: Sample file upload specification in `examples/create_raindrop_file.py` is now correct.\n',
    'author': 'Peter Borocz',
    'author_email': 'peter.borocz+raindrop-io-py@google.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/PBorocz/raindrop-io-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Abstract data types to support Raindrop CLI."""
import os
import sys
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, TypeVar

from rich import print

from raindropiopy.api import (
    API,
    Collection,
    CollectionRef,
    SystemCollection,
    Tag,
    User,
)
from raindropiopy.cli.models.spinner import Spinner

RaindropState = TypeVar(
    "RaindropState",
)  # In py3.11, we'll be able to do 'from typing import Self' instead


@dataclass
class RaindropState:
    """Encapsulate all aspects for current state of the Raindrop environment."""

    api: API
    user: User
    created: datetime = datetime.utcnow()
    collections: list[Collection] = None
    tags: list[str] = None
    refreshed: datetime = None

    def get_collection_titles(
        self,
        exclude_unsorted=False,
        casefold=False,
    ) -> list[str]:
        """Return a sorted list of Collection titles, with or without 'Unsorted' and either casefolded or not."""
        if exclude_unsorted:
            titles = [
                collection.title
                for collection in self.collections
                if collection.id > 0 and collection.title
            ]
        else:
            titles = [
                collection.title for collection in self.collections if collection.title
            ]

        if casefold:
            titles = [title.casefold() for title in titles if title]

        return sorted(titles)

    def find_collection(self, title: str) -> Optional[Collection]:
        """Find the actual Collection object with the title provided."""
        for collection in self.collections:
            if title.casefold() == collection.title.casefold():
                return collection
        return None

    def find_collection_by_id(self, id: int) -> Optional[Collection]:
        """Find the actual Collection object with the *id* provided."""
        for collection in self.collections:
            if collection.id == id:
                return collection
        return None

    def refresh(self) -> bool:
        """Refresh the current state of this Raindrop environment (ie. current collections and tags available)."""

        def __refresh_collections():
            """Gather and set the *COLLECTIONS* associated with the current API user."""
            # This get's us both "root" and "children" collections:
            collections: list[Collection] = Collection.get_collections(self.api)

            # Add the "Unsorted" system collection (we don't care about "Trash" or "All" here)
            for collection in SystemCollection.get_status(self.api):
                if collection.id == CollectionRef.Unsorted.id:
                    unsorted_collection = Collection(
                        _id=collection.id,
                        count=collection.count,
                        title=collection.title,
                        user=self.user,
                    )
                    collections.append(unsorted_collection)

            # Leave our collections in a presentable order, ie. by Title.
            self.collections = sorted(
                collections,
                key=lambda collection: getattr(collection, "title", ""),
            )

        def __refresh_tags():
            """Gather and set the *TAGS* associated with the current API user."""
            # use set to get rid of potential duplicates
            tags: set[str] = set([tag.tag for tag in Tag.get(self.api)])
            self.tags = list(sorted(tags))

        ################################################################################
        with Spinner("Refreshing Raindrop Status..."):
            __refresh_collections()
            __refresh_tags()
            self.refreshed = datetime.utcnow()

        return True

    @classmethod
    def factory(cls) -> RaindropState:
        """Log into Raindrop and return a new, refreshed RaindropState instance."""
        with Spinner("Logging into Raindrop..."):
            api: API = API(
                os.environ["RAINDROP_TOKEN"],
            )  # Setup our connection to Raindrop
            try:
                user = User.get(api)  # What user are we currently defined "for"?
            except Exception:
                print(
                    "[red]Sorry, unable to get query Raindrop.IO for the user requested!",
                )
                sys.exit(1)
        state = RaindropState(api=api, user=user)
        state.refresh()
        return state
